@@ -1,5 +1,5 @@
 /*
- * Copyright © 2007-2008 Gerd Kohlberger <lowfi@chello.at>
+ * Copyright © 2007-2009 Gerd Kohlberger <lowfi@chello.at>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,12 +16,13 @@
  */
 
 #include <gtk/gtk.h>
-#include <glade/glade.h>
 #include <panel-applet.h>
 #include <panel-applet-gconf.h>
 
 #include "mt-common.h"
 
+#define WID(n) (GTK_WIDGET (gtk_builder_get_object (cd->ui, n)))
+
 #define TANGO_CHAMELEON_DARK  0.305f, 0.603f, 0.023f
 #define TANGO_SCARLETRED_DARK 0.643f, 0.000f, 0.000f
 #define TANGO_ALUMINIUM2_DARK 0.180f, 0.203f, 0.211f
@@ -29,25 +30,27 @@
 typedef struct _CaptureData CaptureData;
 struct _CaptureData {
     PanelApplet *applet;
+    GtkBuilder  *ui;
+    GtkWidget   *area;
 
-    GladeXML  *xml;
-    GtkWidget *area;
-    GtkWidget *prefs;
+    GdkCursor   *blank;
+    GdkCursor   *root;
+    gboolean     pointer_locked;
+    gint         pointer_x;
+    gint         pointer_y;
+    gint         center_x;
+    gint         center_y;
+    gboolean     vertical;
 
-    GdkCursor *null_cursor;
-    gboolean   pointer_locked;
-    gint       pointer_x;
-    gint       pointer_y;
-    gboolean   vertical;
-
-    gint  size;
-    gint  cap_button;
-    guint cap_mask;
-    gint  rel_button;
-    guint rel_mask;
+    /* options */
+    gint         size;
+    gint         cap_button;
+    guint        cap_mask;
+    gint         rel_button;
+    guint        rel_mask;
 };
 
-static void fini_capture_data (CaptureData *cd);
+static void capture_data_free (CaptureData *cd);
 
 static void
 capture_preferences (BonoboUIComponent *component,
@@ -56,7 +59,7 @@ capture_preferences (BonoboUIComponent *component,
 {
     CaptureData *cd = data;
 
-    gtk_window_present (GTK_WINDOW (cd->prefs));
+    gtk_window_present (GTK_WINDOW (WID ("capture_preferences")));
 }
 
 static void
@@ -73,7 +76,7 @@ capture_about (BonoboUIComponent *component, gpointer data, const char *cname)
 {
     CaptureData *cd = data;
 
-    gtk_window_present (GTK_WINDOW (glade_xml_get_widget (cd->xml, "about")));
+    gtk_window_present (GTK_WINDOW (WID ("about")));
 }
 
 static const BonoboUIVerb menu_verb[] = {
@@ -83,39 +86,53 @@ static const BonoboUIVerb menu_verb[] = {
     BONOBO_UI_VERB_END
 };
 
-/* grab pointer and update lock state */
-static gboolean
-lock_pointer (CaptureData *cd, guint32 time)
+static void
+lock_pointer (CaptureData *cd)
 {
-    GdkGrabStatus stat;
+    GdkWindow *win;
+    gint x, y, w, h;
 
-    stat = gdk_pointer_grab (cd->area->window,
-			     FALSE,
-			     GDK_BUTTON_PRESS_MASK,
-			     cd->area->window,
-			     cd->null_cursor,
-			     time);
+    /* set invisible cursor */
+    win = gdk_screen_get_root_window (gtk_widget_get_screen (cd->area));
+    cd->root = gdk_window_get_cursor (win);
+    gdk_window_set_cursor (win, cd->blank);
 
-    if (stat != GDK_GRAB_SUCCESS)
-	return FALSE;
+    /* calculate center position */
+    win = gtk_widget_get_window (cd->area);
+    gdk_drawable_get_size (win, &w, &h);
+    gdk_window_get_origin (win, &x, &y);
+    cd->center_x = x + (w >> 1);
+    cd->center_y = y + (h >> 1);
 
+    /* update state */
     cd->pointer_locked = TRUE;
     gtk_widget_queue_draw (cd->area);
-
-    return TRUE;
 }
 
 static void
-unlock_pointer (CaptureData *cd, guint32 time)
+unlock_pointer (CaptureData *cd)
 {
+    GdkWindow *root;
+    GdkCursor *cursor;
+
     /* move pointer to the position where it was locked */
     gdk_display_warp_pointer (gdk_display_get_default (),
 			      gtk_widget_get_screen (cd->area),
 			      cd->pointer_x,
 			      cd->pointer_y);
 
-    gdk_pointer_ungrab (time);
+    /* restore cursor */
+    root = gdk_screen_get_root_window (gtk_widget_get_screen (cd->area));
+    if (cd->root) {
+	gdk_window_set_cursor (root, cd->root);
+    }
+    else {
+	cursor = gdk_cursor_new (GDK_LEFT_PTR);
+	gdk_window_set_cursor (root, cursor);
+	gdk_cursor_unref (cursor);
+    }
 
+    /* update state */
     cd->pointer_locked = FALSE;
     gtk_widget_queue_draw (cd->area);
 }
@@ -124,7 +141,7 @@ unlock_pointer (CaptureData *cd, guint32 time)
 static gboolean
 area_entered (GtkWidget *widget, GdkEventCrossing *cev, gpointer data)
 {
-    CaptureData *cd = (CaptureData *) data;
+    CaptureData *cd = data;
 
     if (cev->mode != GDK_CROSSING_NORMAL)
 	return FALSE;
@@ -134,15 +151,31 @@ area_entered (GtkWidget *widget, GdkEventCrossing *cev, gpointer data)
 
     cd->pointer_x = cev->x_root;
     cd->pointer_y = cev->y_root;
+    lock_pointer (cd);
 
-    return lock_pointer (cd, cev->time);
+    return FALSE;
+}
+
+static gboolean
+area_left (GtkWidget *widget, GdkEventCrossing *cev, gpointer data)
+{
+    CaptureData *cd = data;
+
+    if (cd->pointer_locked) {
+	/* move pointer back to center */
+	gdk_display_warp_pointer (gdk_display_get_default (),
+				  gtk_widget_get_screen (widget),
+				  cd->center_x,
+				  cd->center_y);
+    }
+    return FALSE;
 }
 
 /* change lock states */
 static gboolean
 area_button_press (GtkWidget *widget, GdkEventButton *bev, gpointer data)
 {
-    CaptureData *cd = (CaptureData *) data;
+    CaptureData *cd = data;
 
     if (bev->button == cd->cap_button &&
 	(bev->state & cd->cap_mask) == cd->cap_mask &&
@@ -150,13 +183,13 @@ area_button_press (GtkWidget *widget, GdkEventButton *bev, gpointer data)
 
 	cd->pointer_x = bev->x_root;
 	cd->pointer_y = bev->y_root;
-	lock_pointer (cd, bev->time);
+	lock_pointer (cd);
     }
     else if (bev->button == cd->rel_button &&
 	     (bev->state & cd->rel_mask) == cd->rel_mask &&
 	     cd->pointer_locked) {
 
-	unlock_pointer (cd, bev->time);
+	unlock_pointer (cd);
     }
     else if (!cd->pointer_locked && bev->button != 1) {
 	g_signal_stop_emission_by_name (widget, "button_press_event");
@@ -217,7 +250,7 @@ render_text (cairo_t *cr, CaptureData *cd)
 static gboolean
 area_exposed (GtkWidget *widget, GdkEventExpose *eev, gpointer data)
 {
-    CaptureData *cd = (CaptureData *) data;
+    CaptureData *cd = data;
     cairo_t *cr;
     gdouble w, h;
 
@@ -271,28 +304,28 @@ update_orientation (CaptureData *cd, PanelAppletOrient orient)
 static void
 applet_unrealize (GtkWidget *widget, gpointer data)
 {
-    fini_capture_data ((CaptureData *) data);
+    capture_data_free (data);
 }
 
 static void
 applet_orient_changed (PanelApplet *applet, guint orient, gpointer data)
 {
-    update_orientation ((CaptureData *) data, orient);
+    update_orientation (data, orient);
 }
 
 static void
 about_response (GtkButton *button, gint response, gpointer data)
 {
-    CaptureData *cd = (CaptureData *) data;
+    CaptureData *cd = data;
 
-    gtk_widget_hide (glade_xml_get_widget (cd->xml, "about"));
+    gtk_widget_hide (WID ("about"));
 }
 
 /* preferences dialog callbacks */
 static void
 prefs_size_changed (GtkSpinButton *spin, gpointer data)
 {
-    CaptureData *cd = (CaptureData *) data;
+    CaptureData *cd = data;
 
     cd->size = gtk_spin_button_get_value_as_int (spin);
     panel_applet_gconf_set_int (cd->applet, "size", cd->size, NULL);
@@ -303,7 +336,9 @@ prefs_size_changed (GtkSpinButton *spin, gpointer data)
 static void
 prefs_closed (GtkButton *button, gpointer data)
 {
-    gtk_widget_hide (((CaptureData *) data)->prefs);
+    CaptureData *cd = data;
+
+    gtk_widget_hide (WID ("capture_preferences"));
 }
 
 static void
@@ -316,7 +351,7 @@ prefs_help (GtkButton *button, gpointer data)
 static void
 prefs_cap_button (GtkSpinButton *spin, gpointer data)
 {
-    CaptureData *cd = (CaptureData *) data;
+    CaptureData *cd = data;
 
     cd->cap_button = gtk_spin_button_get_value_as_int (spin);
     panel_applet_gconf_set_int (cd->applet,
@@ -328,7 +363,7 @@ prefs_cap_button (GtkSpinButton *spin, gpointer data)
 static void
 prefs_cap_alt (GtkToggleButton *toggle, gpointer data)
 {
-    CaptureData *cd = (CaptureData *) data;
+    CaptureData *cd = data;
 
     cd->cap_mask ^= GDK_MOD1_MASK;
     panel_applet_gconf_set_bool (cd->applet,
@@ -340,7 +375,7 @@ prefs_cap_alt (GtkToggleButton *toggle, gpointer data)
 static void
 prefs_cap_shift (GtkToggleButton *toggle, gpointer data)
 {
-    CaptureData *cd = (CaptureData *) data;
+    CaptureData *cd = data;
 
     cd->cap_mask ^= GDK_SHIFT_MASK;
     panel_applet_gconf_set_bool (cd->applet,
@@ -352,7 +387,7 @@ prefs_cap_shift (GtkToggleButton *toggle, gpointer data)
 static void
 prefs_cap_ctrl (GtkToggleButton *toggle, gpointer data)
 {
-    CaptureData *cd = (CaptureData *) data;
+    CaptureData *cd = data;
 
     cd->cap_mask ^= GDK_CONTROL_MASK;
     panel_applet_gconf_set_bool (cd->applet,
@@ -364,7 +399,7 @@ prefs_cap_ctrl (GtkToggleButton *toggle, gpointer data)
 static void
 prefs_rel_button (GtkSpinButton *spin, gpointer data)
 {
-    CaptureData *cd = (CaptureData *) data;
+    CaptureData *cd = data;
 
     cd->rel_button = gtk_spin_button_get_value_as_int (spin);
     panel_applet_gconf_set_int (cd->applet,
@@ -376,7 +411,7 @@ prefs_rel_button (GtkSpinButton *spin, gpointer data)
 static void
 prefs_rel_alt (GtkToggleButton *toggle, gpointer data)
 {
-    CaptureData *cd = (CaptureData *) data;
+    CaptureData *cd = data;
 
     cd->rel_mask ^= GDK_MOD1_MASK;
     panel_applet_gconf_set_bool (cd->applet,
@@ -388,7 +423,7 @@ prefs_rel_alt (GtkToggleButton *toggle, gpointer data)
 static void
 prefs_rel_shift (GtkToggleButton *toggle, gpointer data)
 {
-    CaptureData *cd = (CaptureData *) data;
+    CaptureData *cd = data;
 
     cd->rel_mask ^= GDK_SHIFT_MASK;
     panel_applet_gconf_set_bool (cd->applet,
@@ -400,7 +435,7 @@ prefs_rel_shift (GtkToggleButton *toggle, gpointer data)
 static void
 prefs_rel_ctrl (GtkToggleButton *toggle, gpointer data)
 {
-    CaptureData *cd = (CaptureData *) data;
+    CaptureData *cd = data;
 
     cd->rel_mask ^= GDK_CONTROL_MASK;
     panel_applet_gconf_set_bool (cd->applet,
@@ -413,66 +448,73 @@ static gboolean
 init_preferences (CaptureData *cd)
 {
     GtkWidget *w;
+    GError *error = NULL;
 
-    cd->xml = glade_xml_new (DATADIR "/pointer-capture-applet.glade",
-			     NULL, NULL);
-    if (!cd->xml)
+    cd->ui = gtk_builder_new ();
+    gtk_builder_add_from_file (cd->ui,
+			       DATADIR "/pointer-capture-applet.ui",
+			       &error);
+    if (error) {
+	g_print ("%s\n", error->message);
+	g_error_free (error);
 	return FALSE;
+    }
 
-    cd->prefs = glade_xml_get_widget (cd->xml, "capture_preferences");
-    g_signal_connect (cd->prefs, "delete-event",
+    g_signal_connect (WID ("capture_preferences"), "delete-event",
 		      G_CALLBACK (gtk_widget_hide_on_delete), NULL);
-
-    w = glade_xml_get_widget (cd->xml, "close");
-    g_signal_connect (w, "clicked",
+    g_signal_connect (WID ("close"), "clicked",
 		      G_CALLBACK (prefs_closed), cd);
-
-    w = glade_xml_get_widget (cd->xml, "help");
-    g_signal_connect (w, "clicked",
+    g_signal_connect (WID ("help"), "clicked",
 		      G_CALLBACK (prefs_help), NULL);
 
-    w = glade_xml_get_widget (cd->xml, "size");
+    w = WID ("size");
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (w), cd->size);
     g_signal_connect (w, "value_changed",
 		      G_CALLBACK (prefs_size_changed), cd);
 
     /* capture modifier signals */
-    w = glade_xml_get_widget (cd->xml, "cap_button");
+    w = WID ("cap_button");
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (w), cd->cap_button);
     g_signal_connect (w, "value_changed",
 		      G_CALLBACK (prefs_cap_button), cd);
-    w = glade_xml_get_widget (cd->xml, "cap_alt");
+
+    w = WID ("cap_alt");
     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (w),
 				  (cd->cap_mask & GDK_MOD1_MASK));
     g_signal_connect (w, "toggled",
 		      G_CALLBACK (prefs_cap_alt), cd);
-    w = glade_xml_get_widget (cd->xml, "cap_shift");
+
+    w = WID ("cap_shift");
     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (w),
 				  (cd->cap_mask & GDK_SHIFT_MASK));
     g_signal_connect (w, "toggled",
 		      G_CALLBACK (prefs_cap_shift), cd);
-    w = glade_xml_get_widget (cd->xml, "cap_ctrl");
+
+    w = WID ("cap_ctrl");
     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (w),
 				  (cd->cap_mask & GDK_CONTROL_MASK));
     g_signal_connect (w, "toggled",
 		      G_CALLBACK (prefs_cap_ctrl), cd);
 
     /* release modifier signals */
-    w = glade_xml_get_widget (cd->xml, "rel_button");
+    w = WID ("rel_button");
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (w), cd->rel_button);
     g_signal_connect (w, "value_changed",
 		      G_CALLBACK (prefs_rel_button), cd);
-    w = glade_xml_get_widget (cd->xml, "rel_alt");
+
+    w = WID ("rel_alt");
     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (w),
 				  (cd->rel_mask & GDK_MOD1_MASK));
     g_signal_connect (w, "toggled",
 		      G_CALLBACK (prefs_rel_alt), cd);
-    w = glade_xml_get_widget (cd->xml, "rel_shift");
+
+    w = WID ("rel_shift");
     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (w),
 				  (cd->rel_mask & GDK_SHIFT_MASK));
     g_signal_connect (w, "toggled",
 		      G_CALLBACK (prefs_rel_shift), cd);
-    w = glade_xml_get_widget (cd->xml, "rel_ctrl");
+
+    w = WID ("rel_ctrl");
     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (w),
 				  (cd->rel_mask & GDK_CONTROL_MASK));
     g_signal_connect (w, "toggled",
@@ -482,26 +524,27 @@ init_preferences (CaptureData *cd)
 }
 
 static CaptureData *
-init_capture_data (PanelApplet *applet)
+capture_data_init (PanelApplet *applet)
 {
     CaptureData *cd;
 
     cd = g_slice_new0 (CaptureData);
-    if (!cd)
-	return NULL;
-
     cd->applet = applet;
     cd->size = 100;
     cd->rel_button = 1;
+    cd->blank = gdk_cursor_new (GDK_BLANK_CURSOR);
 
     return cd;
 }
 
 static void
-fini_capture_data (CaptureData *cd)
+capture_data_free (CaptureData *cd)
 {
-    if (cd->null_cursor)
-	gdk_cursor_unref (cd->null_cursor);
+    if (cd->blank)
+	gdk_cursor_unref (cd->blank);
+
+    if (cd->ui)
+	g_object_unref (cd->ui);
 
     g_slice_free (CaptureData, cd);
 }
@@ -511,37 +554,17 @@ fill_applet (PanelApplet *applet)
 {
     CaptureData *cd;
     GtkWidget *about;
-    GdkPixmap *bmp0 = NULL;
-    gchar char0[] = { 0x0 };
-    GdkColor c0 = { 0, 0, 0, 0 };
     AtkObject *obj;
 
-    cd = init_capture_data (applet);
-    if (!cd)
-	return FALSE;
-
-    /* invisible cursor */
-    bmp0 = gdk_bitmap_create_from_data (NULL, char0, 1, 1);
-    if (!bmp0) {
-	g_free (cd);
-	return FALSE;
-    }
-
-    cd->null_cursor = gdk_cursor_new_from_pixmap (bmp0, bmp0, &c0, &c0, 0, 0);
-    g_object_unref (bmp0);
-
-    if (!cd->null_cursor) {
-	g_free (cd);
-	return FALSE;
-    }
-
     bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
     bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
     textdomain (GETTEXT_PACKAGE);
 
-    g_set_application_name (_("Pointer Capture Applet"));
+    g_set_application_name (_("Pointer Capture"));
     gtk_window_set_default_icon_name (MT_ICON_NAME);
 
+    cd = capture_data_init (applet);
+
     /* gconf settings */
     panel_applet_add_preferences (applet,
 				  "/schemas/apps/pointer-capture",
@@ -567,7 +590,7 @@ fill_applet (PanelApplet *applet)
 
     /* preferences dialog */
     if (!init_preferences (cd)) {
-	fini_capture_data (cd);
+	capture_data_free (cd);
 	return FALSE;
     }
 
@@ -575,10 +598,13 @@ fill_applet (PanelApplet *applet)
     cd->area = gtk_drawing_area_new ();
     gtk_widget_add_events (cd->area,
 			   GDK_ENTER_NOTIFY_MASK |
+			   GDK_LEAVE_NOTIFY_MASK |
 			   GDK_BUTTON_PRESS_MASK);
 
     g_signal_connect (cd->area, "enter-notify-event",
 		      G_CALLBACK (area_entered), cd);
+    g_signal_connect (cd->area, "leave-notify-event",
+		      G_CALLBACK (area_left), cd);
     g_signal_connect (cd->area, "expose-event",
 		      G_CALLBACK (area_exposed), cd);
     g_signal_connect (cd->area, "button-press-event",
@@ -588,7 +614,7 @@ fill_applet (PanelApplet *applet)
     gtk_widget_show (cd->area);
 
     /* about dialog */
-    about = glade_xml_get_widget (cd->xml, "about");
+    about = WID ("about");
     g_object_set (about, "version", VERSION, NULL);
 
     g_signal_connect (about, "delete-event",
@@ -615,7 +641,6 @@ fill_applet (PanelApplet *applet)
     atk_object_set_description (obj, _("Temporarily lock the mouse pointer"));
 
     gtk_container_add (GTK_CONTAINER (applet), cd->area);
-    gtk_container_set_border_width (GTK_CONTAINER (applet), 1);
     gtk_widget_show (GTK_WIDGET (applet));
 
     return TRUE;
